// -*- C++ -*-

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@nist.gov. 
 */

#include <oofconfig.h>
#include <iostream>
#include "common/cdebug.h"
#include "common/ooferror.h"
#include "common/oofswigruntime.h"
#include "common/pythonlock.h"
#include "common/tostring.h"


// Some trivial constructors are here instead of in ooferror.h so that
// debugging lines can be added without recompiling everything else.

ErrProgrammingError::ErrProgrammingError(const std::string &f, int l)
  : ErrProgrammingErrorBase<ErrProgrammingError>(f, l)
{}

ErrProgrammingError::ErrProgrammingError(const std::string &m,
					 const std::string &f, int l)
  : ErrProgrammingErrorBase<ErrProgrammingError>(m, f, l)
{}

const std::string *ErrBadIndex::summary() const {
  const std::string *sum = ErrProgrammingErrorBase<ErrBadIndex>::summary();
  const std::string *equiv = new std::string(*sum  + " Bad Index: "
					     + tostring(badindex));
  delete sum;
  return equiv;
}

const std::string &ErrProgrammingError::classname() const {
  static std::string s("ErrProgrammingError");
  return s;
}

const std::string &ErrPyProgrammingError::classname() const {
  static std::string s("ErrPyProgrammingError");
  return s;
}

const std::string &ErrResourceShortage::classname() const {
  static std::string s("ErrResourceShortage");
  return s;
}

const std::string &ErrBoundsError::classname() const {
  static std::string s("ErrBoundsError");
  return s;
}

const std::string &ErrBadIndex::classname() const {
  static std::string s("ErrBadIndex");
  return s;
}

const std::string &ErrUserError::classname() const {
  static std::string s("ErrUserError");
  return s;
}

const std::string &ErrSetupError::classname() const {
  static std::string s("ErrSetupError");
  return s;
}

const std::string &ErrInterrupted::classname() const {
  static std::string s("ErrInterrupted");
  return s;
}

const std::string &ErrNoProgress::classname() const {
  static std::string s("ErrNoProgress");
  return s;
}

const std::string &ErrDataFileError::classname() const {
  static std::string s("ErrDataFileError");
  return s;
}

const std::string &ErrWarning::classname() const {
  static std::string s("ErrWarning");
  return s;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Function to convert a Python exception into a C++ exception.  This
// in effect is the inverse of the swig exception typemap.  The two of
// them together allow an exception to be batted back and forth
// between C++ and Python until somebody decides to catch it.

// pythonErrorRelay() should be called whenever C++ detects that a
// Python exception has been raised in a Python API call.  It converts
// the Python exception to a C++ exception if possible.  Otherwise it
// just throws PythonError, which is caught by the swig exception
// typemap.

void pythonErrorRelay() {
  // pythonErrorRelay should be called inside a
  // PYTHON_THREAD_BEGIN_BLOCK/PYTHON_THREAD_END_BLOCK block. 

  PyObject *ptype;
  PyObject *pvalue;
  PyObject *ptraceback;
  // PyErr_Fetch returns new references to its arguments.  It also
  // clears the Python error state.
  PyErr_Fetch(&ptype, &pvalue, &ptraceback);
  if(ptype) {
    // Get the ErrError class, so that we can compare error
    // types on the Python side.
    static PyObject *pyErrError = 0;
    if(!pyErrError) {
      PyObject *module = PyImport_ImportModule((char*) 
					       "ooflib.SWIG.common.ooferror");
      pyErrError = PyObject_GetAttrString(module, (char*) "ErrError");
      assert(pyErrError!= 0);
      Py_XDECREF(module);
    }
    if(PyObject_IsInstance(ptype, pyErrError)) {
      // The Python exception is a swigged ErrError -- ie, it's one of
      // ours.  Extract the C++ object and raise it as a new
      // exception.
      const ErrError *ee;
      PyObject *cexcept = PyObject_GetAttrString(pvalue, "cexcept");
      assert(cexcept != 0);
      SWIG_ConvertPtr(pvalue, (void**) &ee, ((SwigPyObject*) cexcept)->ty, 0);
      Py_XDECREF(ptraceback);
      Py_XDECREF(ptype);
      Py_XDECREF(cexcept);
      // Don't decref pvalue! It will destroy *ee.
      ee->throw_self();
    }
    else {
      // The error is a native Python exception.  Since PyErr_Fetch
      // cleared the error state, we have to call PyErr_Restore.  The
      // error will be caught again when C++ returns to Python, at
      // which point the swig exception typemap will handle it.
      // PyErr_Restore steals references to its arguments, so we don't
      // have to decref here.

      // Mysterious crashes are sometimes due to python errors that
      // don't propagate properly, perhaps because they're occuring
      // within the thread handling machinery.  Uncommenting these
      // lines can help to debug them:
      // std::cerr << "pythonErrorRelay: " << repr(ptype) << std::endl;
      // std::cerr << "pythonErrorRelay: " << repr(pvalue) << std::endl;

      PyErr_Restore(ptype, pvalue, ptraceback);

      throw PythonError();
    }
  } // end if ptype
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// This isn't the real initialization for pyconverter, this is just
// here to make the linker happy.  The real initialization happens in
// the function below at import time.
PyObject *ErrError::pyconverter(0);

// Initializes the static pyconverter member of the ErrError base
// class -- called from the SWIG'd Python file at ooferror import
// time.  Errors which occur prior to this import will not be handled
// correctly.
void pyErrorInit(PyObject *converter) {
  PYTHON_THREAD_BEGIN_BLOCK;
  Py_XINCREF(converter);
  ErrError::pyconverter = converter;
}


